from abc import ABC
from typing import TypeVar, Iterable, TypeAlias, overload
from dataclasses import field, dataclass

from kirin.ir import Block, Region, SSAValue, Statement
from kirin.lattice import BoundedLattice
from kirin.worklist import WorkList
from kirin.interp.base import BaseInterpreter, InterpreterMeta
from kirin.interp.frame import Frame
from kirin.interp.value import Successor, YieldValue, ReturnValue, SpecialValue
from kirin.interp.exceptions import InterpreterError

ResultType = TypeVar("ResultType", bound=BoundedLattice)
WorkListType = TypeVar("WorkListType", bound=WorkList[Successor])

AbsIntResultType: TypeAlias = (
    tuple[ResultType, ...] | None | ReturnValue[ResultType] | YieldValue[ResultType]
)


@dataclass
class AbstractFrame(Frame[ResultType]):
    """Interpreter frame for abstract interpreter.

    This frame is used to store the state of the abstract interpreter.
    It contains the worklist of successors to be processed.
    """

    worklist: WorkList[Successor[ResultType]] = field(default_factory=WorkList)
    visited: dict[Block, set[Successor[ResultType]]] = field(default_factory=dict)


AbstractFrameType = TypeVar("AbstractFrameType", bound=AbstractFrame)

# TODO: support custom loop termination heurestics, e.g. max iteration, etc.
# currently we may end up in infinite loop


class AbstractInterpreterMeta(InterpreterMeta):
    pass


@dataclass
class AbstractInterpreter(
    BaseInterpreter[AbstractFrameType, ResultType],
    ABC,
    metaclass=AbstractInterpreterMeta,
):
    """Abstract interpreter for the IR.

    This is a base class for implementing abstract interpreters for the IR.
    It provides a framework for implementing abstract interpreters given a
    bounded lattice type.

    The abstract interpreter is a forward dataflow analysis that computes
    the abstract values for each SSA value in the IR. The abstract values
    are computed by evaluating the statements in the IR using the abstract
    lattice operations.

    The abstract interpreter is implemented as a worklist algorithm. The
    worklist contains the successors of the current block to be processed.
    The abstract interpreter processes each successor by evaluating the
    statements in the block and updating the abstract values in the frame.

    The abstract interpreter provides hooks for customizing the behavior of
    the interpreter.
    The [`prehook_succ`][kirin.interp.abstract.AbstractInterpreter.prehook_succ] and
    [`posthook_succ`][kirin.interp.abstract.AbstractInterpreter.posthook_succ] methods
    can be used to perform custom actions before and after processing a successor.
    """

    lattice: type[BoundedLattice[ResultType]] = field(init=False)
    """lattice type for the abstract interpreter.
    """

    def __init_subclass__(cls) -> None:
        if ABC in cls.__bases__:
            return super().__init_subclass__()

        if not hasattr(cls, "lattice"):
            raise TypeError(
                f"missing lattice attribute in abstract interpreter class {cls}"
            )
        cls.void = cls.lattice.bottom()
        super().__init_subclass__()

    def prehook_succ(self, frame: AbstractFrameType, succ: Successor):
        """Hook called before processing a successor.

        This method can be used to perform custom actions before processing
        a successor. It is called before evaluating the statements in the block.

        Args:
            frame: The current frame of the interpreter.
            succ: The successor to be processed.
        """
        return

    def posthook_succ(self, frame: AbstractFrameType, succ: Successor):
        """Hook called after processing a successor.

        This method can be used to perform custom actions after processing
        a successor. It is called after evaluating the statements in the block.

        Args:
            frame: The current frame of the interpreter.
            succ: The successor that was processed.
        """
        return

    def should_exec_stmt(self, stmt: Statement) -> bool:
        """This method can be used to control which statements are executed
        during the abstract interpretation. By default, all statements are
        executed.

        This method is useful when one wants to skip certain statements
        during the abstract interpretation and is certain that the skipped
        statements do not affect the final result. This would allow saving
        computation time and memory by not evaluating the skipped statements
        and their results.

        Args:
            stmt: The statement to be executed.

        Returns:
            True if the statement should be executed, False otherwise.
        """
        return True

    def set_values(
        self,
        frame: AbstractFrameType,
        ssa: Iterable[SSAValue],
        results: Iterable[ResultType],
    ):
        """Set the abstract values for the given SSA values in the frame.

        This method is used to customize how the abstract values are set in
        the frame. By default, the abstract values are set directly in the
        frame.
        """
        frame.set_values(ssa, results)

    def eval_recursion_limit(
        self, frame: AbstractFrameType
    ) -> tuple[AbstractFrameType, ResultType]:
        return frame, self.lattice.bottom()

    def run_ssacfg_region(
        self, frame: AbstractFrameType, region: Region, args: tuple[ResultType, ...]
    ) -> tuple[ResultType, ...] | None | ReturnValue[ResultType]:
        result = None
        frame.worklist.append(Successor(region.blocks[0], *args))
        while (succ := frame.worklist.pop()) is not None:
            if succ.block in frame.visited:
                if succ in frame.visited[succ.block]:
                    continue
            else:
                frame.visited[succ.block] = set()
            self.prehook_succ(frame, succ)
            block_result = self.run_succ(frame, succ)
            if len(frame.visited[succ.block]) < 128:
                frame.visited[succ.block].add(succ)
            else:
                continue

            if isinstance(block_result, Successor):
                raise InterpreterError(
                    "unexpected successor, successors should be in worklist"
                )

            result = self.join_results(result, block_result)
            self.posthook_succ(frame, succ)

        if isinstance(result, YieldValue):
            return result.values
        return result

    def run_succ(
        self, frame: AbstractFrameType, succ: Successor
    ) -> SpecialValue[ResultType]:
        frame.current_block = succ.block
        self.set_values(frame, succ.block.args, succ.block_args)
        for stmt in succ.block.stmts:
            if self.should_exec_stmt(stmt) is False:
                continue

            frame.current_stmt = stmt
            stmt_results = self.eval_stmt(frame, stmt)
            if isinstance(stmt_results, tuple):
                self.set_values(frame, stmt._results, stmt_results)
            elif stmt_results is None:
                continue  # empty result
            else:  # terminate
                return stmt_results
        return None

    @overload
    def join_results(self, old: None, new: None) -> None: ...
    @overload
    def join_results(
        self, old: ReturnValue[ResultType], new: ReturnValue[ResultType]
    ) -> ReturnValue[ResultType]: ...
    @overload
    def join_results(
        self, old: YieldValue[ResultType], new: YieldValue[ResultType]
    ) -> YieldValue[ResultType]: ...
    @overload
    def join_results(
        self, old: tuple[ResultType], new: tuple[ResultType]
    ) -> tuple[ResultType]: ...
    @overload
    def join_results(
        self, old: AbsIntResultType[ResultType], new: AbsIntResultType[ResultType]
    ) -> AbsIntResultType[ResultType]: ...

    def join_results(
        self,
        old: AbsIntResultType[ResultType],
        new: AbsIntResultType[ResultType],
    ) -> AbsIntResultType[ResultType]:
        if old is None:
            return new
        elif new is None:
            return old

        if isinstance(old, ReturnValue) and isinstance(new, ReturnValue):
            return ReturnValue(old.value.join(new.value))
        elif isinstance(old, YieldValue) and isinstance(new, YieldValue):
            return YieldValue(
                tuple(
                    old_val.join(new_val)
                    for old_val, new_val in zip(old.values, new.values)
                )
            )
        elif isinstance(old, tuple) and isinstance(new, tuple):
            return tuple(old_val.join(new_val) for old_val, new_val in zip(old, new))
        else:
            return None



from dataclasses import dataclass

from kirin import ir
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.dialects.py.assign import TypeAssert


@dataclass
class InlineTypeAssert(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, TypeAssert):
            return RewriteResult()

        if node.got.type.is_subseteq(node.expected):
            node.got.type = node.got.type.meet(node.expected)
            node.result.replace_by(node.got)
            node.delete()
            return RewriteResult(has_done_something=True)
        return RewriteResult()



import typing

from kirin import lowering

from . import stmts
from .runtime import IList

ElemT = typing.TypeVar("ElemT")
OutElemT = typing.TypeVar("OutElemT")
LenT = typing.TypeVar("LenT")
ResultT = typing.TypeVar("ResultT")

# NOTE: we use Callable here to make nested function work.


@typing.overload
def range(stop: int) -> IList[int, typing.Any]: ...


@typing.overload
def range(start: int, stop: int) -> IList[int, typing.Any]: ...


@typing.overload
def range(start: int, stop: int, step: int) -> IList[int, typing.Any]: ...


@lowering.wraps(stmts.Range)
def range(start: int, stop: int, step: int) -> IList[int, typing.Any]: ...


@lowering.wraps(stmts.Map)
def map(
    fn: typing.Callable[[ElemT], OutElemT],
    collection: IList[ElemT, LenT] | list[ElemT],
) -> IList[OutElemT, LenT]: ...


@lowering.wraps(stmts.Foldr)
def foldr(
    fn: typing.Callable[[ElemT, OutElemT], OutElemT],
    collection: IList[ElemT, LenT] | list[ElemT],
    init: OutElemT,
) -> OutElemT: ...


@lowering.wraps(stmts.Foldl)
def foldl(
    fn: typing.Callable[[OutElemT, ElemT], OutElemT],
    collection: IList[ElemT, LenT] | list[ElemT],
    init: OutElemT,
) -> OutElemT: ...


@lowering.wraps(stmts.Scan)
def scan(
    fn: typing.Callable[[OutElemT, ElemT], tuple[OutElemT, ResultT]],
    collection: IList[ElemT, LenT] | list[ElemT],
    init: OutElemT,
) -> tuple[OutElemT, IList[ResultT, LenT]]: ...


@lowering.wraps(stmts.ForEach)
def for_each(
    fn: typing.Callable[[ElemT], typing.Any],
    collection: IList[ElemT, LenT] | list[ElemT],
) -> None: ...



from dataclasses import dataclass

from .abc import Attribute
from .types import Union, Generic, Literal, PyClass, TypeVar, TypeAttribute

@dataclass
class _TypeAttribute(Attribute):
    def is_subseteq_Union(self, other: Union) -> bool: ...
    def is_subseteq_Literal(self, other: Literal) -> bool: ...
    def is_subseteq_TypeVar(self, other: TypeVar) -> bool: ...
    def is_subseteq_PyClass(self, other: PyClass) -> bool: ...
    def is_subseteq_Generic(self, other: Generic) -> bool: ...
    def is_subseteq_fallback(self, other: TypeAttribute) -> bool: ...



from kirin.prelude import basic_no_opt
from kirin.rewrite import Walk, Fixpoint
from kirin.analysis import const
from kirin.rewrite.dce import DeadCodeElimination
from kirin.rewrite.fold import ConstantFold
from kirin.rewrite.wrap_const import WrapConst


@basic_no_opt
def foldable(x: int) -> int:
    y = 1
    b = y + 2
    c = y + b
    d = c + 4
    return d + x


def test_dce():
    before = foldable(1)
    const_prop = const.Propagate(foldable.dialects)
    frame, _ = const_prop.run_analysis(foldable)
    Fixpoint(Walk(WrapConst(frame))).rewrite(foldable.code)
    fold = ConstantFold()
    Fixpoint(Walk(fold)).rewrite(foldable.code)

    foldable.code.print()
    dce = DeadCodeElimination()
    Fixpoint(Walk(dce)).rewrite(foldable.code)
    foldable.code.print()

    after = foldable(1)

    assert before == after



from attrs import Food, Serving
from dialect import dialect

from kirin import ir, types
from kirin.decl import info, statement


@statement(dialect=dialect)
class NewFood(ir.Statement):
    name = "new_food"
    traits = frozenset({ir.Pure(), ir.FromPythonCall()})
    type: str = info.attribute(types.String)
    result: ir.ResultValue = info.result(types.PyClass(Food))


@statement(dialect=dialect)
class Cook(ir.Statement):
    traits = frozenset({ir.FromPythonCall()})
    target: ir.SSAValue = info.argument(types.PyClass(Food))
    amount: ir.SSAValue = info.argument(types.Int)
    result: ir.ResultValue = info.result(types.PyClass(Serving))


@statement(dialect=dialect)
class Eat(ir.Statement):
    traits = frozenset({ir.FromPythonCall()})
    target: ir.SSAValue = info.argument(types.PyClass(Serving))


@statement(dialect=dialect)
class Nap(ir.Statement):
    traits = frozenset({ir.FromPythonCall()})


@statement(dialect=dialect)
class RandomBranch(ir.Statement):
    name = "random_br"
    traits = frozenset({ir.IsTerminator()})
    cond: ir.SSAValue = info.argument(types.Bool)
    then_arguments: tuple[ir.SSAValue, ...] = info.argument()
    else_arguments: tuple[ir.SSAValue, ...] = info.argument()
    then_successor: ir.Block = info.block()
    else_successor: ir.Block = info.block()



# type: ignore
from group import food
from stmts import Eat, Nap, Cook, NewFood
from recept import FeeAnalysis

from emit import EmitReceptMain
from interp import FoodMethods as FoodMethods
from lattice import AtLeastXItem
from rewrite import NewFoodAndNap
from kirin.rewrite import Walk


@food
def main(x: int):
    food = NewFood(type="burger")  # (1)!
    serving = Cook(food, x)  # (2)!
    Eat(serving)  # (3)!
    Nap()  # (4)!

    return x + 1  # (5)!


main.print()


@food
def main2(x: int):
    def some_closure(food, amount):
        Cook(food, amount + 1)
        Nap()

    fish = NewFood(type="fish")
    chicken = NewFood(type="chicken")
    potatoes = NewFood(type="potatoes")

    fish_serving = Cook(fish, 12 + x)
    chicken_serving = Cook(chicken, 10 + x)
    potatoes_serving = Cook(potatoes, 8)
    Eat(fish_serving)
    Nap()

    some_closure(fish, 1 + 1)
    if x > 1:
        Eat(chicken_serving)
    else:
        Eat(potatoes_serving)
    return x + 1


main2.code.print()
main2(1)  # execute the function
# for i in range(10):
#     print("iteration", i)
#     main(i)  # now eat a random food!


# 2. simple rewrite:
@food
def main3():

    sandwich = NewFood(type="sandwich")
    chips = NewFood(type="chips")

    sandwich_serving = Cook(sandwich, 2)
    chips_serving = Cook(chips, 10)

    Eat(sandwich_serving)
    Eat(chips_serving)


main3.print()
Walk(NewFoodAndNap()).rewrite(main3.code)
main3.print()


# 3. simple analysis example:
@food
def analysis_demo(x: int):

    burger = NewFood(type="burger")
    salad = NewFood(type="salad")

    burger_serving = Cook(burger, 12 + x)
    salad_serving = Cook(salad, 10 + x)

    Eat(burger_serving)
    Eat(salad_serving)
    Nap()

    Eat(burger_serving)
    Nap()

    Eat(burger_serving)
    Nap()

    return x


fee_analysis = FeeAnalysis(analysis_demo.dialects)
results, expect = fee_analysis.run_analysis(
    analysis_demo, args=(AtLeastXItem(data=10),)
)
print(results.entries)
print(fee_analysis.nap_count)
analysis_demo.print(analysis=results.entries)


emitter = EmitReceptMain()
emitter.recept_analysis_result = results.entries

emitter.run(analysis_demo, ("",))
print(emitter.get_output())



# type: ignore
# This file is generated by gen.py
import math as pymath

from kirin.prelude import basic
from kirin.dialects import math


@basic
def acos_func(x):
    return math.acos(x)


def test_acos():
    truth = pymath.acos(0.42)
    assert (acos_func(0.42) - truth) < 1e-6


@basic
def asin_func(x):
    return math.asin(x)


def test_asin():
    truth = pymath.asin(0.42)
    assert (asin_func(0.42) - truth) < 1e-6


@basic
def asinh_func(x):
    return math.asinh(x)


def test_asinh():
    truth = pymath.asinh(0.42)
    assert (asinh_func(0.42) - truth) < 1e-6


@basic
def atan_func(x):
    return math.atan(x)


def test_atan():
    truth = pymath.atan(0.42)
    assert (atan_func(0.42) - truth) < 1e-6


@basic
def atan2_func(y, x):
    return math.atan2(y, x)


def test_atan2():
    truth = pymath.atan2(0.42, 0.42)
    assert (atan2_func(0.42, 0.42) - truth) < 1e-6


@basic
def atanh_func(x):
    return math.atanh(x)


def test_atanh():
    truth = pymath.atanh(0.42)
    assert (atanh_func(0.42) - truth) < 1e-6


@basic
def ceil_func(x):
    return math.ceil(x)


def test_ceil():
    truth = pymath.ceil(0.42)
    assert (ceil_func(0.42) - truth) < 1e-6


@basic
def copysign_func(x, y):
    return math.copysign(x, y)


def test_copysign():
    truth = pymath.copysign(0.42, 0.42)
    assert (copysign_func(0.42, 0.42) - truth) < 1e-6


@basic
def cos_func(x):
    return math.cos(x)


def test_cos():
    truth = pymath.cos(0.42)
    assert (cos_func(0.42) - truth) < 1e-6


@basic
def cosh_func(x):
    return math.cosh(x)


def test_cosh():
    truth = pymath.cosh(0.42)
    assert (cosh_func(0.42) - truth) < 1e-6


@basic
def degrees_func(x):
    return math.degrees(x)


def test_degrees():
    truth = pymath.degrees(0.42)
    assert (degrees_func(0.42) - truth) < 1e-6


@basic
def erf_func(x):
    return math.erf(x)


def test_erf():
    truth = pymath.erf(0.42)
    assert (erf_func(0.42) - truth) < 1e-6


@basic
def erfc_func(x):
    return math.erfc(x)


def test_erfc():
    truth = pymath.erfc(0.42)
    assert (erfc_func(0.42) - truth) < 1e-6


@basic
def exp_func(x):
    return math.exp(x)


def test_exp():
    truth = pymath.exp(0.42)
    assert (exp_func(0.42) - truth) < 1e-6


@basic
def expm1_func(x):
    return math.expm1(x)


def test_expm1():
    truth = pymath.expm1(0.42)
    assert (expm1_func(0.42) - truth) < 1e-6


@basic
def fabs_func(x):
    return math.fabs(x)


def test_fabs():
    truth = pymath.fabs(0.42)
    assert (fabs_func(0.42) - truth) < 1e-6


@basic
def floor_func(x):
    return math.floor(x)


def test_floor():
    truth = pymath.floor(0.42)
    assert (floor_func(0.42) - truth) < 1e-6


@basic
def fmod_func(x, y):
    return math.fmod(x, y)


def test_fmod():
    truth = pymath.fmod(0.42, 0.42)
    assert (fmod_func(0.42, 0.42) - truth) < 1e-6


@basic
def gamma_func(x):
    return math.gamma(x)


def test_gamma():
    truth = pymath.gamma(0.42)
    assert (gamma_func(0.42) - truth) < 1e-6


@basic
def isfinite_func(x):
    return math.isfinite(x)


def test_isfinite():
    truth = pymath.isfinite(0.42)
    assert (isfinite_func(0.42) - truth) < 1e-6


@basic
def isinf_func(x):
    return math.isinf(x)


def test_isinf():
    truth = pymath.isinf(0.42)
    assert (isinf_func(0.42) - truth) < 1e-6


@basic
def isnan_func(x):
    return math.isnan(x)


def test_isnan():
    truth = pymath.isnan(0.42)
    assert (isnan_func(0.42) - truth) < 1e-6


@basic
def lgamma_func(x):
    return math.lgamma(x)


def test_lgamma():
    truth = pymath.lgamma(0.42)
    assert (lgamma_func(0.42) - truth) < 1e-6


@basic
def log10_func(x):
    return math.log10(x)


def test_log10():
    truth = pymath.log10(0.42)
    assert (log10_func(0.42) - truth) < 1e-6


@basic
def log1p_func(x):
    return math.log1p(x)


def test_log1p():
    truth = pymath.log1p(0.42)
    assert (log1p_func(0.42) - truth) < 1e-6


@basic
def log2_func(x):
    return math.log2(x)


def test_log2():
    truth = pymath.log2(0.42)
    assert (log2_func(0.42) - truth) < 1e-6


@basic
def pow_func(x, y):
    return math.pow(x, y)


def test_pow():
    truth = pymath.pow(0.42, 0.42)
    assert (pow_func(0.42, 0.42) - truth) < 1e-6


@basic
def radians_func(x):
    return math.radians(x)


def test_radians():
    truth = pymath.radians(0.42)
    assert (radians_func(0.42) - truth) < 1e-6


@basic
def remainder_func(x, y):
    return math.remainder(x, y)


def test_remainder():
    truth = pymath.remainder(0.42, 0.42)
    assert (remainder_func(0.42, 0.42) - truth) < 1e-6


@basic
def sin_func(x):
    return math.sin(x)


def test_sin():
    truth = pymath.sin(0.42)
    assert (sin_func(0.42) - truth) < 1e-6


@basic
def sinh_func(x):
    return math.sinh(x)


def test_sinh():
    truth = pymath.sinh(0.42)
    assert (sinh_func(0.42) - truth) < 1e-6


@basic
def sqrt_func(x):
    return math.sqrt(x)


def test_sqrt():
    truth = pymath.sqrt(0.42)
    assert (sqrt_func(0.42) - truth) < 1e-6


@basic
def tan_func(x):
    return math.tan(x)


def test_tan():
    truth = pymath.tan(0.42)
    assert (tan_func(0.42) - truth) < 1e-6


@basic
def tanh_func(x):
    return math.tanh(x)


def test_tanh():
    truth = pymath.tanh(0.42)
    assert (tanh_func(0.42) - truth) < 1e-6


@basic
def trunc_func(x):
    return math.trunc(x)


def test_trunc():
    truth = pymath.trunc(0.42)
    assert (trunc_func(0.42) - truth) < 1e-6


@basic
def ulp_func(x):
    return math.ulp(x)


def test_ulp():
    truth = pymath.ulp(0.42)
    assert (ulp_func(0.42) - truth) < 1e-6



from typing import Callable
from dataclasses import field, dataclass

from kirin.ir import Block, Region, Statement
from kirin.worklist import WorkList
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.ir.nodes.base import IRNode


@dataclass
class Walk(RewriteRule):
    """Walk through the IR nodes and apply a rewrite rule.

    The walk will apply the rewrite rule to each node in the IR tree in a depth-first order.

    ### Parameters
    - `map`: The rewrite rule to apply.
    - `reverse`: Whether to traverse the IR tree in reverse order. Default is `False`.
    - `region_first`: Whether to traverse the regions before the blocks. Default is `False`.
    """

    rule: RewriteRule
    worklist: WorkList[IRNode] = field(default_factory=WorkList)
    skip: Callable[[IRNode], bool] = field(default=lambda _: False)

    # options
    reverse: bool = field(default=False)
    region_first: bool = field(default=False)

    def rewrite(self, node: IRNode) -> RewriteResult:
        # NOTE: because the rewrite pass may mutate the node
        # thus we need to save the list of nodes to be processed
        # first before we start processing them
        self.populate_worklist(node)
        has_done_something = False
        subnode = self.worklist.pop()
        while subnode is not None:
            result = self.rule.rewrite(subnode)
            if result.terminated:
                return result

            if result.has_done_something:
                has_done_something = True
            subnode = self.worklist.pop()
        return RewriteResult(has_done_something=has_done_something)

    def populate_worklist(self, node: IRNode) -> None:
        if self.skip(node):
            return

        if isinstance(node, Statement):
            self.populate_worklist_Statement(node)
        elif isinstance(node, Region):
            self.populate_worklist_Region(node)
        elif isinstance(node, Block):
            self.populate_worklist_Block(node)
        else:
            raise NotImplementedError(f"populate_worklist_{node.__class__.__name__}")

    def populate_worklist_Statement(self, node: Statement) -> None:
        if self.region_first:
            self.worklist.append(node)

        if node.regions:
            for region in reversed(node.regions) if not self.reverse else node.regions:
                self.populate_worklist(region)

        if not self.region_first:
            self.worklist.append(node)

    def populate_worklist_Region(self, node: Region) -> None:
        self.worklist.append(node)
        if node.blocks:
            for block in reversed(node.blocks) if not self.reverse else node.blocks:
                self.populate_worklist(block)

    def populate_worklist_Block(self, node: Block) -> None:
        self.worklist.append(node)
        stmt = node.last_stmt
        while stmt is not None:
            self.populate_worklist(stmt)
            stmt = stmt.prev_stmt



import sys
import typing
import inspect
import textwrap
from types import ModuleType

# from typing import TYPE_CHECKING, Generic, TypeVar, Callable, ParamSpec
from dataclasses import field, dataclass

from rich.console import Console

from kirin.ir.traits import HasSignature, CallableStmtInterface
from kirin.ir.exception import ValidationError
from kirin.ir.nodes.stmt import Statement
from kirin.print.printer import Printer
from kirin.ir.attrs.types import Generic
from kirin.print.printable import Printable

if typing.TYPE_CHECKING:
    from kirin.ir.group import DialectGroup

Param = typing.ParamSpec("Param")
RetType = typing.TypeVar("RetType")


@dataclass
class Method(Printable, typing.Generic[Param, RetType]):
    mod: ModuleType | None  # ref
    py_func: typing.Callable[Param, RetType] | None  # ref
    sym_name: str
    arg_names: list[str]
    dialects: "DialectGroup"  # own
    code: Statement  # own, the corresponding IR, a func.func usually
    # values contained if closure
    fields: tuple = field(default_factory=tuple)  # own
    file: str = ""
    lineno_offset: int = 0
    inferred: bool = False
    """if typeinfer has been run on this method
    """

    def __hash__(self) -> int:
        return id(self)

    def __call__(self, *args: Param.args, **kwargs: Param.kwargs) -> RetType:
        from kirin.interp.concrete import Interpreter

        if len(args) + len(kwargs) != len(self.arg_names) - 1:
            raise ValueError("Incorrect number of arguments")
        # NOTE: multi-return values will be wrapped in a tuple for Python
        interp = Interpreter(self.dialects)
        return interp.run(self, args=args, kwargs=kwargs)

    @property
    def args(self):
        """Return the arguments of the method. (excluding self)"""
        return tuple(arg for arg in self.callable_region.blocks[0].args[1:])

    @property
    def arg_types(self):
        """Return the types of the arguments of the method. (excluding self)"""
        return tuple(arg.type for arg in self.args)

    @property
    def self_type(self):
        """Return the type of the self argument of the method."""
        trait = self.code.get_present_trait(HasSignature)
        signature = trait.get_signature(self.code)
        return Generic(Method, Generic(tuple, *signature.inputs), signature.output)

    @property
    def callable_region(self):
        trait = self.code.get_present_trait(CallableStmtInterface)
        return trait.get_callable_region(self.code)

    @property
    def return_type(self):
        trait = self.code.get_present_trait(HasSignature)
        return trait.get_signature(self.code).output

    def __repr__(self) -> str:
        return f'Method("{self.sym_name}")'

    def print_impl(self, printer: Printer) -> None:
        return printer.print(self.code)

    def similar(self, dialects: typing.Optional["DialectGroup"] = None):
        return Method(
            self.mod,
            self.py_func,
            self.sym_name,
            self.arg_names,
            dialects or self.dialects,
            self.code.from_stmt(self.code, regions=[self.callable_region.clone()]),
            self.fields,
            self.file,
            self.inferred,
        )

    def verify(self) -> None:
        """verify the method body.

        This will raise a ValidationError if the method body is not valid.
        """
        try:
            self.code.verify()
        except ValidationError as e:
            self.__postprocess_validation_error(e)

    def verify_type(self) -> None:
        """verify the method type.

        This will raise a ValidationError if the method type is not valid.
        """
        # NOTE: verify the method body
        self.verify()

        try:
            self.code.verify_type()
        except ValidationError as e:
            self.__postprocess_validation_error(e)

    def __postprocess_validation_error(self, e: ValidationError):
        console = Console(force_terminal=True, force_jupyter=False, file=sys.stderr)
        printer = Printer(console=console)
        # NOTE: populate the printer with the method body
        with printer.string_io():
            printer.print(self.code)

        with printer.string_io() as io:
            printer.print(e.node)
            node_str = io.getvalue()

        node_str = "\n".join(["  " + each_line for each_line in node_str.splitlines()])
        msg = f"Incorrect IR:\n    {node_str}\n\n"
        if e.node.source and self.py_func:  # print hint if we have a source
            msg += "\nSource Traceback:"
            source = textwrap.dedent(inspect.getsource(self.py_func))
            msg += e.node.source.error_hint(
                source.splitlines(),
                e,
                file=self.file,
            )
        else:
            msg += "\nNo source available"
            msg += "\nError: " + str(e)
        raise ValidationError(e.node, msg, *e.args, help=e.help) from e



import inspect
from abc import ABC
from typing import TYPE_CHECKING, TypeVar, ClassVar
from dataclasses import dataclass

from kirin.interp.base import BaseInterpreter
from kirin.interp.impl import ImplDef, AttributeImplDef

if TYPE_CHECKING:
    from kirin.ir import Attribute
    from kirin.interp.base import BaseInterpreter
    from kirin.interp.impl import Signature, MethodFunction, AttributeFunction


InterpreterType = TypeVar("InterpreterType", bound="BaseInterpreter")
ValueType = TypeVar("ValueType")


@dataclass
class MethodTable(ABC):
    """Base class to define lookup tables for interpreting code for IR nodes in a dialect."""

    table: ClassVar[dict["Signature", "MethodFunction"]]
    """Lookup table for interpreting code for IR nodes in a dialect."""
    attribute: ClassVar[dict[type["Attribute"], "AttributeFunction"]]
    """Lookup table for interpreting code for IR attributes in a dialect."""

    def __init_subclass__(cls) -> None:
        # init the subclass first
        super().__init_subclass__()
        cls.table = {}
        for _, value in inspect.getmembers(cls):
            if isinstance(value, ImplDef):
                for sig in value.signature:
                    cls.table[sig] = value.impl

        cls.attribute = {}
        for _, value in inspect.getmembers(cls):
            if isinstance(value, AttributeImplDef):
                cls.attribute[value.signature] = value.impl



from kirin import types
from kirin.prelude import basic


def test_global_hint():
    @basic
    def main(xs: types.Float) -> None:  # type: ignore
        return None

    assert main.code.signature.inputs[0] == types.Float  # type: ignore
    assert main.code.signature.output == types.NoneType  # type: ignore



from kirin.exception import NoPythonStackTrace


class BuildError(NoPythonStackTrace):
    """Base class for all dialect lowering errors."""

    def __init__(self, *msgs: object, help: str | None = None):
        super().__init__(*msgs)
        self.help = help



from kirin import ir, types
from kirin.interp import Frame, Interpreter, MethodTable, impl
from kirin.dialects.py.len import Len
from kirin.dialects.py.binop import Add

from .stmts import Map, New, Push, Scan, Foldl, Foldr, Range, ForEach
from .runtime import IList
from ._dialect import dialect


@dialect.register
class IListInterpreter(MethodTable):

    @impl(Range)
    def _range(self, interp, frame: Frame, stmt: Range):
        return (IList(range(*frame.get_values(stmt.args)), elem=types.Int),)

    @impl(New)
    def new(self, interp, frame: Frame, stmt: New):
        return (IList(list(frame.get_values(stmt.values)), elem=stmt.elem_type),)

    @impl(Len, types.PyClass(IList))
    def len(self, interp, frame: Frame, stmt: Len):
        return (len(frame.get(stmt.value).data),)

    @impl(Add, types.PyClass(IList), types.PyClass(IList))
    def add(self, interp, frame: Frame, stmt: Add):
        lhs, rhs = frame.get(stmt.lhs), frame.get(stmt.rhs)
        return (IList(lhs.data + rhs.data, elem=lhs.elem.join(rhs.elem)),)

    @impl(Push)
    def push(self, interp, frame: Frame, stmt: Push):
        lst = frame.get(stmt.lst)
        return (IList(lst.data + [frame.get(stmt.value)], elem=lst.elem),)

    @impl(Map)
    def map(self, interp: Interpreter, frame: Frame, stmt: Map):
        fn: ir.Method = frame.get(stmt.fn)
        coll: IList = frame.get(stmt.collection)
        ret = []
        for elem in coll.data:
            # NOTE: assume fn has been type checked
            _, item = interp.run_method(fn, (elem,))
            ret.append(item)
        return (IList(ret, elem=fn.return_type),)

    @impl(Scan)
    def scan(self, interp: Interpreter, frame: Frame, stmt: Scan):
        fn: ir.Method = frame.get(stmt.fn)
        init = frame.get(stmt.init)
        coll: IList = frame.get(stmt.collection)

        carry = init
        ys = []
        for elem in coll.data:
            # NOTE: assume fn has been type checked
            _, (carry, y) = interp.run_method(fn, (carry, elem))
            ys.append(y)

        if (
            isinstance(fn.return_type, types.Generic)
            and fn.return_type.is_subseteq(types.Tuple)
            and len(fn.return_type.vars) == 2
        ):
            return ((carry, IList(ys, fn.return_type.vars[1])),)
        else:
            return ((carry, IList(ys, types.Any)),)

    @impl(Foldr)
    def foldr(self, interp: Interpreter, frame: Frame, stmt: Foldr):
        return self.fold(interp, frame, stmt, reversed(frame.get(stmt.collection).data))

    @impl(Foldl)
    def foldl(self, interp: Interpreter, frame: Frame, stmt: Foldl):
        return self.fold(interp, frame, stmt, frame.get(stmt.collection).data)

    def fold(self, interp: Interpreter, frame: Frame, stmt: Foldr | Foldl, coll):
        fn: ir.Method = frame.get(stmt.fn)
        init = frame.get(stmt.init)

        acc = init
        for elem in coll:
            # NOTE: assume fn has been type checked
            _, acc = interp.run_method(fn, (acc, elem))
        return (acc,)

    @impl(ForEach)
    def for_each(self, interp: Interpreter, frame: Frame, stmt: ForEach):
        fn: ir.Method = frame.get(stmt.fn)
        coll: IList = frame.get(stmt.collection)
        for elem in coll.data:
            # NOTE: assume fn has been type checked
            interp.run_method(fn, (elem,))
        return (None,)



from kirin.prelude import basic_no_opt


def test_simple():
    @basic_no_opt
    def main(x: int):
        for i in range(5):
            x = x + 1
        return x

    assert main.py_func is not None
    assert main.py_func(1) == main(1)


# generate some more complicated loop
def test_nested():
    @basic_no_opt
    def main(x: int):
        for i in range(5):
            for j in range(5):
                x = x + 1
        return x

    assert main.py_func is not None
    assert main.py_func(1) == main(1)


def test_nested2():
    @basic_no_opt
    def main(x: int):
        for i in range(5):
            for j in range(5):
                for k in range(5):
                    x = x + 1
        return x

    assert main.py_func is not None
    assert main.py_func(1) == main(1)



import pytest

from kirin import ir, types, lowering
from kirin.decl import info, statement
from kirin.prelude import basic

dialect = ir.Dialect("dummy")


@statement(dialect=dialect)
class DummyStmt(ir.Statement):
    name = "dummy"
    traits = frozenset({lowering.FromPythonCall()})
    value: ir.SSAValue = info.argument(types.Int)
    option: ir.PyAttr[str] = info.attribute()
    result: ir.ResultValue = info.result(types.Int)


def test_attribute_lowering():
    @basic.add(dialect)
    def test(x: int):
        return DummyStmt(x, option=ir.PyAttr("attr"))  # type: ignore

    option = test.code.body.blocks[0].stmts.at(0).option  # type: ignore
    assert isinstance(option, ir.PyAttr) and option.data == "attr"

    with pytest.raises(lowering.BuildError):

        @basic.add(dialect)
        def not_working(x: int):
            return DummyStmt(x, option=x)  # type: ignore



from kirin.prelude import basic_no_opt
from kirin.rewrite import Walk, Chain, Fixpoint, WrapConst
from kirin.analysis import const
from kirin.rewrite.dce import DeadCodeElimination
from kirin.rewrite.alias import InlineAlias


@basic_no_opt
def main_simplify_alias(x: int):
    y = x + 1
    z = y
    z2 = z
    return z2


def test_alias_inline():
    constprop = const.Propagate(main_simplify_alias.dialects)
    frame, ret = constprop.run_analysis(main_simplify_alias)
    Fixpoint(Walk(WrapConst(frame))).rewrite(main_simplify_alias.code)
    Fixpoint(Walk(Chain([InlineAlias(), DeadCodeElimination()]))).rewrite(
        main_simplify_alias.code
    )
    assert len(main_simplify_alias.callable_region.blocks[0].stmts) == 3


@basic_no_opt
def simplify_alias_ref_const():
    y = 3
    z = y
    return z


def test_alias_inline2():
    constprop = const.Propagate(simplify_alias_ref_const.dialects)
    frame, _ = constprop.run_analysis(simplify_alias_ref_const)
    Fixpoint(Walk(WrapConst(frame))).rewrite(main_simplify_alias.code)
    Fixpoint(Walk(Chain([InlineAlias(), DeadCodeElimination()]))).rewrite(
        simplify_alias_ref_const.code
    )
    simplify_alias_ref_const.code.print()
    assert len(simplify_alias_ref_const.callable_region.blocks[0].stmts) == 2



"""This module provides access to Python iterables.

This is used to lower Python loops into `cf` dialect.

This module contains the common methods for the Python iterable:

- The `Iter` statement class.
- The `Next` statement class.
- The lowering pass for the iterable.
- The concrete implementation of the iterable.

This dialect maps `iter()` and `next()` calls to the `Iter` and `Next` statements.
"""

from ast import Call

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement

dialect = ir.Dialect("py.iterable")

PyRangeIterType = types.PyClass(type(iter(range(0))))


@statement(dialect=dialect)
class Iter(ir.Statement):
    """This is equivalent to `iter(value)` in Python."""

    traits = frozenset({ir.Pure()})
    value: ir.SSAValue = info.argument(types.Any)
    iter: ir.ResultValue = info.result(types.Any)


@statement(dialect=dialect)
class Next(ir.Statement):
    """This is equivalent to `next(iterable, None)` in Python."""

    iter: ir.SSAValue = info.argument(types.Any)
    value: ir.ResultValue = info.result(types.Any)


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Iter)
    def iter_(self, interp, frame: interp.Frame, stmt: Iter):
        return (iter(frame.get(stmt.value)),)

    @interp.impl(Next)
    def next_(self, interp, frame: interp.Frame, stmt: Next):
        return (next(frame.get(stmt.iter), None),)


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(Iter, types.PyClass(range))
    def iter_(self, interp, frame: interp.Frame, stmt: Iter):
        return (PyRangeIterType,)

    @interp.impl(Next, PyRangeIterType)
    def next_(self, interp, frame: interp.Frame, stmt: Next):
        return (types.Int,)


@dialect.register
class Lowering(lowering.FromPythonAST):

    @lowering.akin(iter)
    def lower_Call_iter(self, state: lowering.State, node: Call) -> lowering.Result:
        if len(node.args) != 1:
            raise lowering.BuildError("iter() takes exactly 1 argument")
        return state.current_frame.push(
            Iter(state.lower(node.args[0]).expect_one()),
        )

    @lowering.akin(next)
    def lower_Call_next(self, state: lowering.State, node: Call) -> lowering.Result:
        if len(node.args) == 2:
            raise lowering.BuildError(
                "next() does not throw StopIteration inside kernel"
            )
        if len(node.args) != 1:
            raise lowering.BuildError("next() takes exactly 1 argument")

        return state.current_frame.push(
            Next(state.lower(node.args[0]).expect_one()),
        )



from typing import TypeVar

from .abc import BoundedLattice

BoundedLatticeType = TypeVar("BoundedLatticeType", bound="BoundedLattice")


class IsSubsetEqMixin(BoundedLattice[BoundedLatticeType]):
    """A special mixin for lattices that provides a default implementation for `is_subseteq`
    by using the visitor pattern. This is useful if the lattice has a lot of different
    subclasses that need to be compared.

    Must be used before `BoundedLattice` in the inheritance chain.
    """

    def is_subseteq(self, other: BoundedLatticeType) -> bool:
        if other is self.top():
            return True
        elif other is self.bottom():
            return False

        method = getattr(
            self,
            "is_subseteq_" + other.__class__.__name__,
            getattr(self, "is_subseteq_fallback", None),
        )
        if method is not None:
            return method(other)
        return False


class SimpleJoinMixin(BoundedLattice[BoundedLatticeType]):
    """A mixin that provides a simple implementation for the join operation."""

    def join(self, other: BoundedLatticeType) -> BoundedLatticeType:
        if self.is_subseteq(other):
            return other
        elif other.is_subseteq(self):
            return self  # type: ignore
        return self.top()


class SimpleMeetMixin(BoundedLattice[BoundedLatticeType]):
    """A mixin that provides a simple implementation for the meet operation."""

    def meet(self, other: BoundedLatticeType) -> BoundedLatticeType:
        if self.is_subseteq(other):
            return self  # type: ignore
        elif other.is_subseteq(self):
            return other
        return self.bottom()



"""Analysis module for kirin.

This module contains the analysis framework for kirin. The analysis framework is
built on top of the interpreter framework. This module provides a set of base classes
and frameworks for implementing compiler analysis passes on the IR.

The analysis framework contains the following modules:

- [`cfg`][kirin.analysis.cfg]: Control flow graph for a given IR.
- [`forward`][kirin.analysis.forward]: Forward dataflow analysis.
- [`callgraph`][kirin.analysis.callgraph]: Call graph for a given IR.
- [`typeinfer`][kirin.analysis.typeinfer]: Type inference analysis.
- [`const`][kirin.analysis.const]: Constants used in the analysis framework.
"""

from kirin.analysis import const as const
from kirin.analysis.cfg import CFG as CFG
from kirin.analysis.forward import (
    Forward as Forward,
    ForwardExtra as ForwardExtra,
    ForwardFrame as ForwardFrame,
)
from kirin.analysis.callgraph import CallGraph as CallGraph
from kirin.analysis.typeinfer import TypeInference as TypeInference



from kirin.interp import Frame, Successor, Interpreter, MethodTable, impl
from kirin.dialects.cf.stmts import Branch, ConditionalBranch
from kirin.dialects.cf.dialect import dialect


@dialect.register
class CfInterpreter(MethodTable):

    @impl(Branch)
    def branch(self, interp: Interpreter, frame: Frame, stmt: Branch):
        return Successor(stmt.successor, *frame.get_values(stmt.arguments))

    @impl(ConditionalBranch)
    def conditional_branch(
        self, interp: Interpreter, frame: Frame, stmt: ConditionalBranch
    ):
        if frame.get(stmt.cond):
            return Successor(
                stmt.then_successor, *frame.get_values(stmt.then_arguments)
            )
        else:
            return Successor(
                stmt.else_successor, *frame.get_values(stmt.else_arguments)
            )



from typing import Callable
from dataclasses import dataclass

from kirin import ir
from kirin.dialects import cf, func
from kirin.rewrite.abc import RewriteRule, RewriteResult

# TODO: use func.Constant instead of kirin.dialects.py.stmts.Constant
from kirin.dialects.py.constant import Constant

# NOTE: this only inlines func dialect


@dataclass
class Inline(RewriteRule):
    heuristic: Callable[[ir.Statement], bool]
    """inline heuristic that determines whether a function should be inlined
    """

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if isinstance(node, func.Invoke):
            return self.rewrite_func_Invoke(node)
        elif isinstance(node, func.Call):
            return self.rewrite_func_Call(node)
        else:
            return RewriteResult()

    def rewrite_func_Call(self, node: func.Call) -> RewriteResult:
        if not isinstance(lambda_stmt := node.callee.owner, func.Lambda):
            return RewriteResult()

        # NOTE: a lambda statement is defined and used in the same scope
        self.inline_call_like(node, tuple(node.args), lambda_stmt.body)
        return RewriteResult(has_done_something=True)

    def rewrite_func_Invoke(self, node: func.Invoke) -> RewriteResult:
        has_done_something = False
        callee = node.callee

        if (
            isinstance(callee, ir.Method)
            and self.heuristic(callee.code)
            and (call_trait := callee.code.get_trait(ir.CallableStmtInterface))
            is not None
        ):
            region = call_trait.get_callable_region(callee.code)
            func_self = Constant(node.callee)
            func_self.result.name = node.callee.sym_name
            func_self.insert_before(node)
            self.inline_call_like(
                node, (func_self.result,) + tuple(arg for arg in node.args), region
            )
            has_done_something = True

        return RewriteResult(has_done_something=has_done_something)

    def inline_call_like(
        self,
        call_like: ir.Statement,
        args: tuple[ir.SSAValue, ...],
        region: ir.Region,
    ):
        """
        Inline a function call-like statement

        Args:
            call_like (ir.Statement): the call-like statement
            args (tuple[ir.SSAValue, ...]): the arguments of the call (first one is the callee)
            region (ir.Region): the region of the callee
        """
        # <stmt>
        # <stmt>
        # <br (a, b, c)>

        # <block (a, b,c)>:
        # <block>:
        # <block>:
        # <br>

        # ^<block>:
        # <stmt>
        # <stmt>

        # 1. we insert the entry block of the callee function
        # 2. we insert the rest of the blocks into the parent region
        # 3.1 if the return is in the entry block, means no control flow,
        #     replace the call results with the return values
        # 3.2 if the return is some of the blocks, means control flow,
        #     split the current block into two, and replace the return with
        #     the branch instruction
        # 4. remove the call
        if not call_like.parent_block:
            return

        if not call_like.parent_region:
            return

        # NOTE: we cannot change region because it may be used elsewhere
        inline_region: ir.Region = region.clone()
        parent_block: ir.Block = call_like.parent_block
        parent_region: ir.Region = call_like.parent_region

        # wrap what's after invoke into a block
        after_block = ir.Block()
        stmt = call_like.next_stmt
        while stmt is not None:
            stmt.detach()
            after_block.stmts.append(stmt)
            stmt = call_like.next_stmt

        for result in call_like.results:
            block_arg = after_block.args.append_from(result.type, result.name)
            result.replace_by(block_arg)

        parent_block_idx = parent_region._block_idx[parent_block]
        entry_block = inline_region.blocks.popfirst()
        idx, block = 0, entry_block
        while block is not None:
            idx += 1

            if block.last_stmt and isinstance(block.last_stmt, func.Return):
                block.last_stmt.replace_by(
                    cf.Branch(
                        arguments=(block.last_stmt.value,),
                        successor=after_block,
                    )
                )

            parent_region.blocks.insert(parent_block_idx + idx, block)
            block = inline_region.blocks.popfirst()

        parent_region.blocks.append(after_block)

        # NOTE: we expect always to have an entry block
        # but we still check for it cuz we are not gonna
        # error for empty regions here.
        if entry_block:
            cf.Branch(
                arguments=args,
                successor=entry_block,
            ).insert_before(call_like)
        call_like.delete()
        return



import numpy as np
import pytest

# composite modules
from aha import gaga, hoho

from kirin.interp import IntepreterExit
from kirin.prelude import basic_no_opt


@basic_no_opt
def ge(x: int) -> int:
    return hoho.foo(x) >= hoho.goo(x) >= gaga.foo(x)


@basic_no_opt
def gt(x: int) -> int:
    return hoho.foo(x) > hoho.goo(x) > gaga.foo(x)


@basic_no_opt
def and_(x: int) -> int:
    return hoho.foo(x) and hoho.goo(x) and gaga.foo(x)


@basic_no_opt
def or_(x: int) -> int:
    return hoho.foo(x) or hoho.goo(x) or gaga.foo(x)


@basic_no_opt
def and_or(x: int) -> int:
    return hoho.foo(x) and hoho.goo(x) or gaga.foo(x)


@basic_no_opt
def not_(x: int) -> int:
    return not (hoho.foo(x) == hoho.goo(x) == gaga.foo(x))


@basic_no_opt
def unary(x: int):
    return -hoho.foo(x), ~hoho.foo(x)


@basic_no_opt
def le(x: int) -> int:
    return hoho.foo(x) <= hoho.goo(x) <= gaga.foo(x)


@basic_no_opt
def lshift(x: int) -> int:
    return x << 1


@basic_no_opt
def rshift(x: int) -> int:
    return x >> 1


@basic_no_opt
def matmul(x, y):
    return x @ y


@basic_no_opt
def bitwise(x: int):
    return x & 1, x | 1, x ^ 1


@basic_no_opt
def invert(x: int) -> int:
    return ~x


@basic_no_opt
def lt(x: int) -> int:
    return hoho.foo(x) < hoho.goo(x) < gaga.foo(x)


@basic_no_opt
def ne(x: int) -> int:
    return hoho.foo(x) != hoho.goo(x) != gaga.foo(x)


@basic_no_opt
def multi_return(x: int):
    return hoho.foo(x), hoho.foo(x) + 1


@basic_no_opt
def add(x: int) -> int:
    return hoho.foo(x) + hoho.goo(x) + gaga.foo(x)


@basic_no_opt
def sub(x: int) -> int:
    return hoho.foo(x) - hoho.goo(x) - gaga.foo(x)


@pytest.mark.parametrize("x", [1, 2, 3])
def test_multi_return(x):
    assert multi_return(x) == (x + 1, x + 2)


@basic_no_opt
def mul(x: int) -> int:
    return hoho.foo(x) * hoho.goo(x) * gaga.foo(x)


@basic_no_opt
def div(x: int):
    return hoho.foo(x) / hoho.goo(x) / gaga.foo(x)


@basic_no_opt
def floordiv(x: int) -> int:
    return hoho.foo(x) // hoho.goo(x) // gaga.foo(x)


@basic_no_opt
def mod(x: int) -> int:
    return hoho.foo(x) % hoho.goo(x) % gaga.foo(x)


@basic_no_opt
def pow(x: int) -> int:
    return hoho.foo(x) ** hoho.goo(x) ** gaga.foo(x)


@basic_no_opt
def eq(x: int) -> int:
    return hoho.foo(x) == hoho.goo(x) == gaga.foo(x)


@pytest.mark.parametrize("x", [1, 2, 3])
def test_add(x):
    if x == 1:
        assert add(x) == (x + 1) + (x + 1) + (x + 1)
    else:
        pytest.raises(IntepreterExit, lambda: add(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_sub(x):
    if x == 1:
        assert sub(x) == (x + 1) - (x + 1) - (x + 1)
    else:
        pytest.raises(IntepreterExit, lambda: sub(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_mul(x):
    if x == 1:
        assert mul(x) == (x + 1) * (x + 1) * (x + 1)
    else:
        pytest.raises(IntepreterExit, lambda: mul(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_div(x):
    if x == 1:
        assert div(x) == (x + 1) / (x + 1) / (x + 1)
    else:
        pytest.raises(IntepreterExit, lambda: div(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_floordiv(x):
    if x == 1:
        assert floordiv(x) == (x + 1) // (x + 1) // (x + 1)
    else:
        pytest.raises(IntepreterExit, lambda: floordiv(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_mod(x):
    if x == 1:
        assert mod(x) == (x + 1) % (x + 1) % (x + 1)
    else:
        pytest.raises(IntepreterExit, lambda: mod(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_pow(x):
    if x == 1:
        assert pow(x) == (x + 1) ** (x + 1) ** (x + 1)
    else:
        pytest.raises(IntepreterExit, lambda: pow(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_eq(x):
    if x == 1:
        assert eq(x) == ((x + 1) == (x + 1) == (x + 1))
    else:
        pytest.raises(IntepreterExit, lambda: eq(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_ne(x):
    if x == 1:
        assert ne(x) == ((x + 1) != (x + 1) != (x + 1))
    else:
        pytest.raises(IntepreterExit, lambda: ne(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_lt(x):
    if x == 1:
        assert lt(x) == ((x + 1) < (x + 1) < (x + 1))
    else:
        pytest.raises(IntepreterExit, lambda: lt(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_le(x):
    if x == 1:
        assert le(x) == ((x + 1) <= (x + 1) <= (x + 1))
    else:
        pytest.raises(IntepreterExit, lambda: le(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_gt(x):
    if x == 1:
        assert gt(x) == ((x + 1) > (x + 1) > (x + 1))
    else:
        pytest.raises(IntepreterExit, lambda: gt(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_ge(x):
    if x == 1:
        assert ge(x) == ((x + 1) >= (x + 1) >= (x + 1))
    else:
        pytest.raises(IntepreterExit, lambda: ge(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_and(x):
    if x == 1:
        assert and_(x) == ((x + 1) and (x + 1) and (x + 1))
    else:
        pytest.raises(IntepreterExit, lambda: and_(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_or(x):
    if x == 1:
        assert or_(x) == ((x + 1) or (x + 1) or (x + 1))
    else:
        pytest.raises(IntepreterExit, lambda: or_(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_and_or(x):
    if x == 1:
        assert and_or(x) == ((x + 1) and (x + 1) or (x + 1))
    else:
        pytest.raises(IntepreterExit, lambda: and_or(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_not(x):
    if x == 1:
        assert not_(x) == (not ((x + 1) == (x + 1) == (x + 1)))
    else:
        pytest.raises(IntepreterExit, lambda: not_(x))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_unary(x):
    assert unary(x) == (-(x + 1), ~(x + 1))


@pytest.mark.parametrize("x", [1, 2, 3])
def test_lshift(x):
    assert lshift(x) == (x << 1)


@pytest.mark.parametrize("x", [1, 2, 3])
def test_rshift(x):
    assert rshift(x) == x >> 1


@pytest.mark.parametrize("x", [1, 2, 3])
def test_bitwise(x):
    assert bitwise(x) == (x & 1, x | 1, x ^ 1)


@pytest.mark.parametrize("x", [1, 2, 3])
def test_invert(x):
    assert invert(x) == ~x


@pytest.mark.parametrize(
    "x, y", [(np.zeros((2, 2)), np.ones((2, 2))), (np.ones((2, 2)), np.ones((2, 2)))]
)
def test_matmul(x, y):
    assert (matmul(x, y) == x @ y).all()



from kirin.decl.base import BaseModifier
from kirin.decl.camel2snake import camel2snake

from ._set_new_attribute import set_new_attribute


class EmitName(BaseModifier):

    def emit_name(self):
        set_new_attribute(self.cls, "name", camel2snake(self.cls.__name__))



from kirin.decl.base import BaseModifier

from ._create_fn import create_fn
from ._set_new_attribute import set_new_attribute


class EmitRepr(BaseModifier):

    def emit_repr(self):
        if "repr" not in self.params or not self.params["repr"]:
            return

        body = [f'ret = "{self.cls.__name__}("']
        for idx, field in enumerate(self.fields):
            if idx > 0:
                body.append('ret += ", "')
            body.append(f'ret += f"{field.name}={{{self._self_name}.{field.name}}}"')
        body.append('ret += ")"')
        body.append("return ret")

        set_new_attribute(
            self.cls,
            "__repr__",
            create_fn(
                "__repr__",
                args=[self._self_name],
                body=body,
                globals=self.globals,
                return_type=str,
            ),
        )



"""Interpreter module for Kirin.

This module contains the interpreter framework for Kirin. The interpreter
framework is used to implement concrete and abstract interpreters for the
IR. The interpreter framework provides a set of classes and interfaces to
implement interpreters for the IR.

The interpreter framework is designed to be extensible and customizable. It
provides a set of base classes and interfaces for implementing concrete and
abstract interpreters:

- [`BaseInterpreter`][kirin.interp.BaseInterpreter]: Base class for implementing concrete interpreters.
- [`AbstractInterpreter`][kirin.interp.AbstractInterpreter]: Base class for implementing abstract interpreters.
- [`Frame`][kirin.interp.Frame]: Base class for interpreter frame.
- [`MethodTable`][kirin.interp.MethodTable]: Method table for registering implementations of statements.
"""

from .base import BaseInterpreter as BaseInterpreter
from .impl import ImplDef as ImplDef, Signature as Signature, impl as impl
from .frame import Frame as Frame, FrameABC as FrameABC
from .table import MethodTable as MethodTable
from .value import (
    Successor as Successor,
    YieldValue as YieldValue,
    ReturnValue as ReturnValue,
    SpecialValue as SpecialValue,
    StatementResult as StatementResult,
)
from .abstract import (
    AbstractFrame as AbstractFrame,
    AbstractInterpreter as AbstractInterpreter,
)
from .concrete import Interpreter as Interpreter
from .exceptions import (
    WrapException as WrapException,
    IntepreterExit as IntepreterExit,
    InterpreterError as InterpreterError,
    FuelExhaustedError as FuelExhaustedError,
)



from __future__ import annotations

import sys
from typing import TYPE_CHECKING
from dataclasses import dataclass

from kirin.exception import CustomStackTrace

if TYPE_CHECKING:
    from .frame import FrameABC
    from .state import InterpreterState


# errors
class InterpreterError(Exception):
    """Generic interpreter error.

    This is the base class for all interpreter errors. Interpreter
    errors will be catched by the interpreter and handled appropriately
    as an error with stack trace (of Kirin, not Python) from the interpreter.
    """

    pass


@dataclass
class WrapException(InterpreterError):
    """A special interpreter error that wraps a Python exception."""

    exception: Exception


@dataclass
class IntepreterExit(CustomStackTrace):
    exception: Exception
    state: InterpreterState

    def print_stacktrace(self) -> None:
        """Print the stacktrace of the interpreter."""
        frame: FrameABC | None = self.state.current_frame
        print(f"{type(self.exception).__name__}: {self.exception}", file=sys.stderr)
        print("Traceback (most recent call last):", file=sys.stderr)
        frames: list[FrameABC] = []
        while frame is not None:
            frames.append(frame)
            frame = frame.parent
        frames.reverse()
        for frame in frames:
            if stmt := frame.current_stmt:
                print("  " + repr(stmt.source), file=sys.stderr)
                print("     " + stmt.print_str(end=""), file=sys.stderr)


class FuelExhaustedError(InterpreterError):
    """An error raised when the interpreter runs out of fuel."""

    pass



# type: ignore

from kirin import ir, types, lowering
from kirin.decl import info, statement
from kirin.prelude import basic, basic_no_opt
from kirin.dialects.py import cmp

dialect = ir.Dialect("dummy2")


@statement(dialect=dialect)
class DummyStmt2(ir.Statement):
    name = "dummy2"
    traits = frozenset({lowering.FromPythonCall()})
    value: ir.SSAValue = info.argument(types.Int)
    option: ir.PyAttr[str] = info.attribute()
    result: ir.ResultValue = info.result(types.Int)


@basic_no_opt.add(dialect)
def unfolable(x: int, y: int):
    def inner():
        DummyStmt2(x, option=ir.PyAttr("hello"))
        DummyStmt2(y, option=ir.PyAttr("hello"))

    return inner


@basic.add(dialect)(fold=True, aggressive=True)
def main():
    x = DummyStmt2(1, option=ir.PyAttr("hello"))
    x = unfolable(x, x)
    return x()


def test_aggressive_pass():
    const_count = 0
    dummy_count = 0
    for stmt in main.callable_region.walk():
        if isinstance(stmt, DummyStmt2):
            dummy_count += 1
        elif stmt.has_trait(ir.ConstantLike):
            const_count += 1
    assert dummy_count == 3
    assert const_count == 2


@basic(fold=True, aggressive=True)
def should_fold():
    return 1 < 2


def test_should_fold():
    for stmt in should_fold.callable_region.walk():
        assert not isinstance(stmt, cmp.Lt)



from kirin import interp
from kirin.emit.julia import EmitJulia, EmitStrFrame

from . import stmts
from ._dialect import dialect


@dialect.register(key="emit.julia")
class JuliaTable(interp.MethodTable):

    @interp.impl(stmts.Add)
    def emit_Add(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.Add):
        return emit.emit_binaryop(frame, "+", stmt.lhs, stmt.rhs, stmt.result)

    @interp.impl(stmts.Sub)
    def emit_Sub(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.Sub):
        return emit.emit_binaryop(frame, "-", stmt.lhs, stmt.rhs, stmt.result)

    @interp.impl(stmts.Mult)
    def emit_Mult(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.Mult):
        return emit.emit_binaryop(frame, "*", stmt.lhs, stmt.rhs, stmt.result)

    @interp.impl(stmts.Div)
    def emit_Div(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.Div):
        return emit.emit_binaryop(frame, "/", stmt.lhs, stmt.rhs, stmt.result)

    @interp.impl(stmts.Mod)
    def emit_Mod(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.Mod):
        return emit.emit_binaryop(frame, "%", stmt.lhs, stmt.rhs, stmt.result)

    @interp.impl(stmts.Pow)
    def emit_Pow(self, emit: EmitJulia, frame: EmitStrFrame, stmt: stmts.Pow):
        return emit.emit_binaryop(frame, "^", stmt.lhs, stmt.rhs, stmt.result)



import numpy as np

from kirin.interp import MethodTable, impl

from .stmts import X, Y, Z, Id
from .dialect import _dialect


@_dialect.register
class PauliMethods(MethodTable):
    X_mat = np.array([[0, 1], [1, 0]])
    Y_mat = np.array([[0, -1j], [1j, 0]])
    Z_mat = np.array([[1, 0], [0, -1]])
    Id_mat = np.array([[1, 0], [0, 1]])

    @impl(X)  # (1)!
    def x(self, interp, frame, stmt: X):
        return (stmt.pre_factor * self.X_mat,)

    @impl(Y)
    def y(self, interp, frame, stmt: Y):
        return (self.Y_mat * stmt.pre_factor,)

    @impl(Z)
    def z(self, interp, frame, stmt: Z):
        return (self.Z_mat * stmt.pre_factor,)

    @impl(Id)
    def id(self, interp, frame, stmt: Id):
        return (self.Id_mat * stmt.pre_factor,)



from kirin.passes import Fold
from kirin.prelude import basic, structural_no_opt


def test_simple_loop():
    @structural_no_opt
    def simple_loop():
        x = 0
        for i in range(2):
            x = x + 1
        return x

    @basic(fold=True)
    def target():
        return 2

    fold = Fold(structural_no_opt)
    fold(simple_loop)
    assert target.callable_region.is_equal(simple_loop.callable_region)



import pytest

from kirin import ir
from kirin.prelude import basic
from kirin.dialects import math


@basic(verify=False, typeinfer=False)
def check_type_err(a, b):
    math.sin(a)
    return math.sin(b)


def test_check_type():
    with pytest.raises(ir.TypeCheckError):
        check_type_err.code.verify_type()
